#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.core.tools
      @file: dict_tools.py
   @created: Thu, 03 May 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from typing import Any, Dict, Iterable, List, Optional, Tuple


def search_dict(root_element: dict, search_path: str, parent_key="", sep=".") -> Optional[Any]:
    """
    TODO
    :param root_element:
    :param search_path:
    :param parent_key:
    :param sep:
    :return:
    """
    found = search_path == parent_key
    el = None
    if not found and isinstance(root_element, dict):
        for key, value in root_element.items():
            if found:
                break
            el_path = parent_key + sep + key if parent_key else key
            if search_path == el_path:
                found, el = True, value
                break
            if isinstance(value, dict):
                found, el = search_dict(value, search_path, el_path, sep=sep)
            elif isinstance(value, list):
                for next_val in value:
                    found, el = search_dict(next_val, search_path, el_path, sep=sep)
            # Skip if the element is a leaf

    return found, el


def flatten_dict(dictionary: dict, parent_key="", sep=".") -> dict:
    """Flatten a dictionary and all it's items.
    :param dictionary: The dictionary to be flattened
    :param parent_key: The parent key name
    :param sep: The separator between keys
    :return:
    """
    flat_dict = {}
    for key, value in dictionary.items():
        new_key = parent_key + sep + key if parent_key else key
        if isinstance(value, dict):
            flat_dict.update(flatten_dict(value, new_key, sep=sep).items())
        else:
            flat_dict.update({new_key: value})

    return flat_dict


def merge(list_of_dicts: List[Dict] | Tuple[Dict]) -> dict:
    """Merge a list of iterables of dicts into a single dictionary.
    :param list_of_dicts: The list of iterables to be merged
    """
    return {a: av for d in list_of_dicts for a, av in d.items()}


def intersect(left_dict: Iterable, right_dict: Iterable) -> Iterable:
    """Intersect two iterables of dicts into a single dictionary.
    :param left_dict: the left side dictionary
    :param right_dict: the right side dictionary
    """
    return type(left_dict)([value for value in left_dict if value in right_dict])


def get_or_default(options: tuple | list, index: int, default_value: Any = None) -> Optional[Any]:
    """Retrieve an item from the options list or default_value if index is out of range
    :param options: The available list of options
    :param index: The index of the item
    :param default_value: The default value if the index is out of range
    """
    return options[index] if index < len(options) else default_value


def get_or_default_by_key(options: dict, key: str, default_value: Any = None) -> Optional[Any]:
    """Retrieve an item from the options list or default_value if key was not found
    :param options: The available list of options
    :param key: The key of the item
    :param default_value: The default value if the index is not found
    """
    return options[key] if key in options else default_value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.modules.cli.vt100
      @file: vt_utils.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from hspylib.core.exception.exceptions import NotATerminalError
from hspylib.core.tools.commons import is_debugging, sysout
from hspylib.modules.cli.vt100.vt_100 import Vt100
from typing import Any, Optional, Tuple

import atexit
import os
import re
import sys
import termios
import tty


def screen_size() -> tuple[int, ...]:
    """Retrieve the size of the terminal"""
    if not sys.stdout.isatty():
        raise NotATerminalError("screen_size:: Requires a terminal (TTY)")

    return tuple(map(int, os.popen("stty size").read().split()))


# Solution taken from:
# https://stackoverflow.com/questions/46651602/determine-the-terminal-cursor-position-with-an-ansi-sequence-in-python-3
def get_cursor_position() -> Optional[Tuple[int, int]]:
    """Get the terminal cursor position"""
    if not sys.stdout.isatty():
        raise NotATerminalError("get_cursor_position:: Requires a terminal (TTY)")

    if is_debugging():
        return 0, 0
    buf = ""
    stdin = sys.stdin.fileno()
    attrs = termios.tcgetattr(stdin)
    try:
        tty.setcbreak(stdin, termios.TCSANOW)
        sys.stdout.write(Vt100.get_cursor_pos())  # Query the terminal
        sys.stdout.flush()
        while True:
            buf += sys.stdin.read(1)
            if buf[-1] == "R":
                break
        matches = re.match(r"^\x1b\[(\d*);(\d*)R", buf)
        groups = matches.groups()
        return int(groups[0]), int(groups[1])
    except AttributeError:
        return None
    finally:
        termios.tcsetattr(stdin, termios.TCSANOW, attrs)


def set_enable_echo(enabled: bool = True) -> None:
    """Enable echo in the terminal
    :param enabled: whether is enabled or not
    """
    if not sys.stdout.isatty():
        raise NotATerminalError("set_enable_echo:: Requires a terminal (TTY)")

    os.popen(f"stty {'echo -raw' if enabled else 'raw -echo min 0'}").read()


def set_auto_wrap(auto_wrap: bool = True) -> None:
    """Set auto-wrap mode in the terminal
    :param auto_wrap: whether auto_wrap is set or not
    """
    sysout(Vt100.set_auto_wrap(auto_wrap), end="")


def set_show_cursor(show_cursor: bool = True) -> None:
    """Show or hide cursor in the terminal
    :param show_cursor: whether to show or hide he cursor
    """
    sysout(Vt100.set_show_cursor(show_cursor), end="")


def erase_line(mode: int = 2) -> None:
    """Erase current line"""
    sysout(f"%EL{mode}%\r", end="")


def save_cursor() -> None:
    """Save cursor position and attributes"""
    sysout(Vt100.save_cursor(), end="")


def clear_screen() -> None:
    """Clear terminal"""
    sysout("%ED2%", end="")
    sysout("%HOM%", end="")


def restore_cursor() -> None:
    """Restore cursor position and attributes"""
    sysout(Vt100.restore_cursor(), end="")


def alternate_screen(enable: bool = True) -> None:
    """Switch to the alternate screen buffer on/off"""
    sysout(f"%SC{'A' if enable else 'M'}%", end="")


def restore_terminal() -> None:
    """Clear the terminal and restore default attributes"""
    set_auto_wrap()
    set_show_cursor()
    set_enable_echo()


def exit_app(exit_code: int = None, frame: Any = None, exit_msg: str = "") -> None:
    """Exit the application. Commonly called by hooked signals"""
    restore_terminal()
    sysout(f"%MOD(0)%" f"{exit_msg if exit_msg else ''}")
    sys.exit(exit_code)


def prepare_render(auto_wrap: bool = False, show_cursor: bool = False):
    """Prepare the terminal for TUI renderization"""
    atexit.register(restore_terminal)
    set_auto_wrap(auto_wrap)
    set_show_cursor(show_cursor)
    clear_screen()
    save_cursor()


if __name__ == '__main__':
    print('1 odiheiwdhewiduhe widuhewiduhewi duhewdi uewhd iweu')
    print('2 odiheiwdhewiduhe widuhewiduhewi duhewdi uewhd iweu')
    print('3 odiheiwdhewiduhe widuhewiduhewi duhewdi uewhd iweu')
    print('4 odiheiwdhewiduhe widuhewiduhewi duhewdi uewhd iweu')
    print('5 odiheiwdhewiduhe widuhewiduhewi duhewdi uewhd iweu')
    clear_screen()
    exit_app(exit_msg='Done')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.core.decorator
      @file: decorators.py
   @created: Tue, 11 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from datetime import datetime, timedelta
from hspylib.core.tools.commons import str_to_bool
from math import ceil
from typing import Any, Callable

import logging as log
import os
import unittest


def integration_test(cls: type):
    """Enable/Disable HsPyLib integration tests"""

    it_disabled = str_to_bool(os.environ.get("HSPYLIB_IT_DISABLED", "True"))
    if it_disabled:
        log.debug("Skipping test: %s", cls.__name__)
        return unittest.skipIf(
            it_disabled, f"Disabled = {it_disabled} :integration tests because it needs real servers running"
        )

    return cls


def temporized(func: Callable):
    """Log the time spent on the Callable invocation."""

    def timedelta_to_str(delta: timedelta) -> str:
        """Convert timedelta object into string"""
        parts = str(delta).split(":")
        h, m, s = parts[0], parts[1], parts[2][:-7]
        ms = int(ceil(int(parts[2][-6:]) / 1000))
        return f"{h:02s}[h] {m:02s}[m] {s:02s}[s] {ms:02d}[ms]"

    def temporized_closure(*args) -> Any:
        """Execute the callable and return"""
        start = datetime.now()
        ret = func(*args)
        log.debug("@@@ [%s] Time elapsed\t%s", func.__name__, timedelta_to_str(datetime.now() - start))
        return ret

    return temporized_closure



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.core.enums
      @file: enumeration.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from enum import auto, Enum
from hspylib.core.preconditions import check_argument, check_not_none
from typing import Any, List, Tuple, TypeVar

E = TypeVar("E", bound="Enumeration")


def composable(cls: type):
    """Make the enumeration class, composable"""
    check_argument(issubclass(cls, Enumeration))
    setattr(cls.__class__, '_CUSTOM', auto())
    return cls


class Enumeration(Enum):
    """Extended enumeration type"""

    @classmethod
    def names(cls) -> List[str]:
        """Return all enumeration names"""
        return list(map(lambda e: e.name, cls))

    @classmethod
    def values(cls) -> List[Any]:
        """Return all enumeration values"""
        return list(map(lambda e: e.value, cls))

    @classmethod
    def value_of(cls, name: str, ignore_case: bool = True) -> E:
        """Create an enumeration provided it's matching name."""
        if ignore_case:
            found = next(filter(lambda en: en.name.upper() == name.upper(), list(cls)), None)
        else:
            found = next(filter(lambda en: en.name == name, list(cls)), None)
        return check_not_none(found, '"{}" name does not correspond to a valid "{}" enum', name, cls.__name__)

    @classmethod
    def of_value(cls, value: Any, ignore_case: bool = False) -> E:
        if ignore_case:
            found = next(filter(lambda en: str(en.value).upper() == str(value).upper(), list(cls)), None)
        else:
            found = next(filter(lambda en: en.value == value, list(cls)), None)
        return check_not_none(found, '"{}" value does not correspond to a valid "{}" enum', value, cls.__name__)

    @classmethod
    def compose(cls, first: E, *others: E) -> E:
        try:
            e = getattr(first.__class__, "_CUSTOM")
            e._value_ = str(first)  # pylint: disable=protected-access
            for other in others:
                e._value_ += str(other)
            return e
        except AttributeError as err:
            raise AttributeError("Composable enumerations must declare a '_CUSTOM' member") from err

    def __str__(self):
        return str(self._value_)

    def __repr__(self):
        return self.name

    def __hash__(self) -> int:
        return hash(self.key)

    def __eq__(self, other: E) -> bool:
        if isinstance(other, self.__class__):
            return self.key == other.key
        return NotImplemented

    def __add__(self, other) -> E:
        return Enumeration.compose(self, other)

    @property
    def key(self) -> Tuple[str, Any]:
        return self.name, self.value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.modules.qt.views
      @file: qt_view.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from abc import ABC
from hspylib.core.preconditions import check_argument, check_state
from hspylib.core.tools.commons import run_dir
from PyQt5 import uic
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QWidget
from typing import Optional, Tuple, Type

import os


class QtView(ABC):
    """TODO"""

    MAIN_QT_VIEW_UI = "main_qt_view.ui"

    @staticmethod
    def load_form(form_file: str, load_dir: str) -> Tuple[Type, Type]:
        """Load the ui form from the .ui file"""

        form_dir = load_dir if load_dir else f"{run_dir()}/resources/forms/"
        check_argument(
            os.path.exists(form_dir) and os.path.isdir(form_dir),
            "Load dir {} does not exist or is not a folder",
            form_dir,
        )
        filepath = f"{form_dir}/{form_file}"
        check_state(
            os.path.exists(filepath) and os.path.isfile(filepath) and filepath.lower().endswith(".ui"),
            "Form file {} does not exist or it is not a valid UI form file",
            form_file,
        )

        return uic.loadUiType(filepath)

    def __init__(self, ui_file: str = MAIN_QT_VIEW_UI, load_dir: str = None, parent: Optional[QWidget] = None):
        ui_clazz, window_clazz = self.load_form(ui_file, load_dir)
        # Must come after the initialization above
        check_state(ui_clazz is not None and window_clazz is not None, "Unable to initialize UI and Window objects")
        self.window, self.ui = window_clazz(), ui_clazz()
        self.ui.setupUi(self.window)
        self.parent = parent

    def show(self) -> None:
        """Show the main widget"""
        self.window.show()

    def set_default_font(self, default_font: QFont = QFont("Courier New", 14)):
        """Set font for all UI components at once"""
        widgets = list(filter(lambda o: hasattr(getattr(self.ui, o), "setFont"), vars(self.ui)))
        list(map(lambda w: getattr(self.ui, w).setFont(default_font), widgets))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.core.enums
      @file: http_code.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from hspylib.core.enums.enumeration import Enumeration
from hspylib.core.preconditions import check_not_none
from hspylib.modules.cli.vt100.vt_color import VtColor


class HttpCode(Enumeration):
    """HTTP status code constants defined by the proposed HTTP/1.1 specification at http://www.ietf.org/rfc/rfc2068.txt.
    Ref:http://www.cs.ait.ac.th/~on/O/oreilly/java-ent/servlet/appc_01.htm
    """

    # fmt: off

    # 1xx Informational Codes
    CONTINUE                            = 100, 'Continue'
    SWITCHING_PROTOCOLS                 = 101, 'Switching Protocols'
    PROCESSING                          = 102, 'Processing'
    EARLY_HINTS                         = 103, 'Early Hints'

    # 2xx Success Codes
    OK                                  = 200, 'OK'
    CREATED                             = 201, 'Created'
    ACCEPTED                            = 202, 'Accepted'
    NON_AUTHORITATIVE_INFORMATION       = 203, 'Non-Authoritative Information'
    NO_CONTENT                          = 204, 'No Content'
    RESET_CONTENT                       = 205, 'Reset Content'
    PARTIAL_CONTENT                     = 206, 'Partial Content'
    MULTI_STATUS                        = 207, 'Multi-Status'
    ALREADY_REPORTED                    = 208, 'Already Reported'
    IM_USED                             = 226, 'IM Used'

    # 3xx Redirection Codes
    MULTIPLE_CHOICES                    = 300, 'Multiple Choices'
    MOVED_PERMANENTLY                   = 301, 'Moved Permanently'
    FOUND                               = 302, 'Found'
    SEE_OTHER                           = 303, 'See Other'
    NOT_MODIFIED                        = 304, 'Not Modified'
    USE_PROXY                           = 305, 'Use Proxy'
    TEMPORARY_REDIRECT                  = 307, 'Temporary Redirect'
    PERMANENT_REDIRECT                  = 308, 'Permanent Redirect'

    # 4xx Client Error Codes
    BAD_REQUEST                         = 400, 'Bad Request'
    UNAUTHORIZED                        = 401, 'Unauthorized'
    PAYMENT_REQUIRED                    = 402, 'Payment Required'
    FORBIDDEN                           = 403, 'Forbidden'
    NOT_FOUND                           = 404, 'Not Found'
    METHOD_NOT_ALLOWED                  = 405, 'Method Not Allowed'
    NOT_ACCEPTABLE                      = 406, 'Not Acceptable'
    PROXY_AUTHENTICATION_REQUIRED       = 407, 'Proxy Authentication Required'
    REQUEST_TIMEOUT                     = 408, 'Request Timeout'
    CONFLICT                            = 409, 'Conflict'
    GONE                                = 410, 'Gone'
    LENGTH_REQUIRED                     = 411, 'Length Required'
    PRECONDITION_FAILED                 = 412, 'Precondition Failed'
    PAYLOAD_TOO_LARGE                   = 413, 'Payload Too Large'
    REQUEST_URI_TOO_LONG                = 414, 'Request-URI Too Long'
    UNSUPPORTED_MEDIA_TYPE              = 415, 'Unsupported Media Type'
    REQUESTED_RANGE_NOT_SATISFIABLE     = 416, 'Requested Range Not Satisfiable'
    EXPECTATION_FAILED                  = 417, 'Expectation Failed'
    MISDIRECTED_REQUEST                 = 421, 'Misdirected Request'
    UNPROCESSABLE_ENTITY                = 422, 'Unprocessable Entity'
    LOCKED                              = 423, 'Locked'
    FAILED_DEPENDENCY                   = 424, 'Failed Dependency'
    TOO_EARLY                           = 425, 'Too Early'
    UPGRADE_REQUIRED                    = 426, 'Upgrade Required'
    PRECONDITION_REQUIRED               = 428, 'Precondition Required'
    TOO_MANY_REQUESTS                   = 429, 'Too Many Requests'
    REQUEST_HEADER_FIELDS_TOO_LARGE     = 431, 'Request Header Fields Too Large'
    CONNECTION_CLOSED_WITHOUT_RESPONSE  = 444, 'Connection Closed Without Response'
    UNAVAILABLE_FOR_LEGAL_REASONS       = 451, 'Unavailable For Legal Reasons'
    CLIENT_CLOSED_REQUEST               = 499, 'Client Closed Request'

    # 5xx Server Error
    INTERNAL_SERVER_ERROR               = 500, 'Internal Server Error'
    NOT_IMPLEMENTED                     = 501, 'Not Implemented'
    BAD_GATEWAY                         = 502, 'Bad Gateway'
    SERVICE_UNAVAILABLE                 = 503, 'Service Unavailable'
    GATEWAY_TIMEOUT                     = 504, 'Gateway Timeout'
    HTTP_VERSION_NOT_SUPPORTED          = 505, 'HTTP Version Not Supported'
    VARIANT_ALSO_NEGOTIATES             = 506, 'Variant Also Negotiates'
    INSUFFICIENT_STORAGE                = 507, 'Insufficient Storage'
    LOOP_DETECTED                       = 508, 'Loop Detected'
    NOT_EXTENDED                        = 510, 'Not Extended'
    NETWORK_AUTHENTICATION_REQUIRED     = 511, 'Network Authentication Required'
    NETWORK_CONNECT_TIMEOUT_ERROR       = 599, 'Network Connect Timeout Error'

    # fmt: on

    @classmethod
    def of(cls, status_code: int) -> "HttpCode":
        found = next(filter(lambda en: en.code == status_code, list(cls)), None)
        return check_not_none(found, '"{}" code does not correspond to a valid "HttpCode"', status_code)

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        if self.is_2xx():
            color = VtColor.GREEN.code
        elif self.is_3xx():
            color = VtColor.YELLOW.code
        elif self.is_4xx() or self.is_5xx():
            color = VtColor.RED.code
        else:
            color = VtColor.WHITE.code

        return f"{color}({self.code}) {self.reason}{VtColor.NC.code}"

    @property
    def code(self) -> int:
        return self.value[0]

    @property
    def reason(self) -> str:
        return self.value[1]

    def is_1xx(self) -> bool:
        return 100 <= self.code < 200

    def is_2xx(self) -> bool:
        return 200 <= self.code < 300

    def is_3xx(self) -> bool:
        return 300 <= self.code < 400

    def is_4xx(self) -> bool:
        return 400 <= self.code < 500

    def is_5xx(self) -> bool:
        return 500 <= self.code < 600



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.modules.qt.promotions
      @file: htableview.py
   @created: Wed, 30 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from hspylib.core.collection_filter import CollectionFilter
from hspylib.core.preconditions import check_argument, check_not_none, check_state
from hspylib.core.tools.text_tools import strip_linebreaks
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QCursor, QPainter, QPaintEvent
from PyQt5.QtWidgets import QAbstractScrollArea, QHeaderView, QMenu, QTableView, QWidget
from typing import Callable, Optional

import os
import pyperclip as clipboard

CB_ACTION = Callable[[], None]


class HTableView(QTableView):
    """Promoted QTableView widget."""

    def __init__(self, parent: Optional[QWidget], placeholder: Optional[str] = None):
        super().__init__(parent)
        self._context_menu_enable = True
        self._copyable = True
        self._clearable = True
        self._deletable = True
        self._placeholder = placeholder or "No data to display"
        self._custom_menu_actions = []
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.context_menu)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.horizontalHeader().setStretchLastSection(True)
        self.verticalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContents)

    def paintEvent(self, event: QPaintEvent) -> None:
        super().paintEvent(event)
        if self.model() is not None and self.model().rowCount() > 0:
            return
        color = self.palette().placeholderText().color()
        painter = QPainter(self.viewport())
        painter.save()
        painter.setPen(color)
        elided_text = self.fontMetrics().elidedText(self._placeholder, Qt.ElideRight, self.viewport().width())
        painter.drawText(self.viewport().rect(), Qt.AlignCenter, elided_text)
        painter.restore()

    def filters(self) -> Optional[CollectionFilter]:
        """Return current data filters"""
        return self.model().filters() if self.model() else None

    def refresh(self) -> None:
        """Synchronize view and model data"""
        self.model().refresh_data()

    def clear(self):
        """Clear the entire table"""
        model = self.model()
        if model:
            model.clear()

    def copy(self) -> None:
        """Copy selected cell into clipboard"""
        sel_model = self.selectionModel()
        if sel_model:
            index_list = sel_model.selectedIndexes()
            text = ""
            last_row = 0
            for index in index_list:
                if len(text) > 0:
                    if last_row == index.row():
                        text += "\t"
                    else:
                        text += os.linesep
                text += strip_linebreaks(str(self.model().column(index)))
                last_row = index.row()
            clipboard.copy(text)

    def delete(self) -> None:
        """Delete selected rows"""
        sel_model = self.selectionModel()
        if sel_model:
            sel_rows = sel_model.selectedRows()
            if sel_rows:
                self.model().remove_rows(sel_rows)

    def context_menu(self) -> None:
        """Display the custom context menu"""
        if not self.is_empty() and self._context_menu_enable:
            ctx_menu = QMenu(self)
            if self._copyable:
                ctx_menu.addAction("Copy Cells", self.copy)
            if self._deletable:
                ctx_menu.addAction("Delete Row", self.delete)
            if self._clearable:
                ctx_menu.addSeparator()
                ctx_menu.addAction("Clear table", self.clear)

            for action in self._custom_menu_actions:
                check_not_none(action)
                check_state(len(action) == 3, f"Invalid custom menu action: {action}")
                check_argument(callable(action[1]), "The action must be callable")
                if action[2]:
                    ctx_menu.addSeparator()
                ctx_menu.addAction(action[0], action[1])

            ctx_menu.exec_(QCursor.pos())

    def add_custom_menu_action(self, item_text: str, action: CB_ACTION, add_separator: bool) -> None:
        """Add a custom menu action item"""
        self._custom_menu_actions.append((item_text, action, add_separator))

    def set_context_menu_enable(self, enabled: bool = True) -> None:
        """Whether context menu is enabled or not"""
        self._context_menu_enable = enabled

    def set_copyable(self, copyable: bool = True) -> None:
        """Whether the widget is copyable or not"""
        self._copyable = copyable

    def set_clearable(self, clearable: bool = True) -> None:
        """Whether the widget is clearable or not"""
        self._clearable = clearable

    def set_deletable(self, deletable: bool = True) -> None:
        """Whether the widget is clearable or not"""
        self._deletable = deletable

    def is_empty(self) -> bool:
        """Whether the table view view has data or not"""
        return not self.model() or self.model().is_empty()



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.162
#
# Package: main.hspylib.modules.application.argparse
"""Package initialization."""

__all__ = [
  'argument_parser', 
  'arguments_builder', 
  'chained_arguments_builder', 
  'options_builder', 
  'parser_action'
]
__version__ = '0.11.162'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.modules.cli
      @file: keyboard.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

import select
import string
import sys
import termios
from typing import List, Optional

import getkey

from hspylib.core.enums.enumeration import Enumeration
from hspylib.core.exception.exceptions import KeyboardInputError, NotATerminalError


# pylint: disable=multiple-statements
class Keyboard(Enumeration):
    """Provides keyboard interaction with the terminal."""

    # fmt: off

    # Control keys
    VK_NONE         = ''; VK_DISABLED = ''

    VK_ESC          = getkey.keys.ESC; VK_ENTER             = getkey.keys.ENTER
    VK_UP           = getkey.keys.UP; VK_DELETE             = getkey.keys.DELETE
    VK_DOWN         = getkey.keys.DOWN; VK_SPACE            = getkey.keys.SPACE
    VK_LEFT         = getkey.keys.LEFT; VK_HOME             = getkey.keys.HOME
    VK_RIGHT        = getkey.keys.RIGHT; VK_END             = getkey.keys.END
    VK_BACKSPACE    = getkey.keys.BACKSPACE; VK_PAGE_UP     = getkey.keys.PAGE_UP
    VK_INSERT       = getkey.keys.INSERT; VK_PAGE_DOWN      = getkey.keys.PAGE_DOWN
    VK_TAB          = '\t'; VK_SHIFT_TAB                    = '\x1b[Z'

    VK_CTRL_A       = '\x01';       VK_CTRL_B = '\x02';       VK_CTRL_C = VK_DISABLED; VK_CTRL_D = '\x04'
    VK_CTRL_E       = '\x05';       VK_CTRL_F = '\x06';       VK_CTRL_G = '\x07';      VK_CTRL_H = '\x08'
    VK_CTRL_I       = VK_DISABLED;  VK_CTRL_J = VK_DISABLED;  VK_CTRL_K = '\x0b';      VK_CTRL_L = '\x0c'
    VK_CTRL_M       = VK_DISABLED;  VK_CTRL_N = '\x0e';       VK_CTRL_O = VK_DISABLED; VK_CTRL_P = '\x10'
    VK_CTRL_Q       = VK_DISABLED;  VK_CTRL_R = '\x12';       VK_CTRL_S = VK_DISABLED; VK_CTRL_T = '\x14'
    VK_CTRL_U       = '\x15';       VK_CTRL_V = VK_DISABLED;  VK_CTRL_W = '\x17';      VK_CTRL_X = '\x18'
    VK_CTRL_Y       = '\x19';       VK_CTRL_Z = VK_DISABLED

    # Letters
    VK_a = 'a'; VK_i = 'i'; VK_q = 'q'; VK_y = 'y'
    VK_A = 'A'; VK_I = 'I'; VK_Q = 'Q'; VK_Y = 'Y'
    VK_b = 'b'; VK_j = 'j'; VK_r = 'r'; VK_z = 'z'
    VK_B = 'B'; VK_J = 'J'; VK_R = 'R'; VK_Z = 'Z'
    VK_c = 'c'; VK_k = 'k'; VK_s = 's'
    VK_C = 'C'; VK_K = 'K'; VK_S = 'S'
    VK_d = 'd'; VK_l = 'l'; VK_t = 't'
    VK_D = 'D'; VK_L = 'L'; VK_T = 'T'
    VK_e = 'e'; VK_m = 'm'; VK_u = 'u'
    VK_E = 'E'; VK_M = 'M'; VK_U = 'U'
    VK_f = 'f'; VK_n = 'n'; VK_v = 'v'
    VK_F = 'F'; VK_N = 'N'; VK_V = 'V'
    VK_g = 'g'; VK_o = 'o'; VK_w = 'w'
    VK_G = 'G'; VK_O = 'O'; VK_W = 'W'
    VK_h = 'h'; VK_p = 'p'; VK_x = 'x'
    VK_H = 'H'; VK_P = 'P'; VK_X = 'X'

    # Numbers
    VK_ZERO         = '0'; VK_FIVE  = '5'
    VK_ONE          = '1'; VK_SIX   = '6'
    VK_TWO          = '2'; VK_SEVEN = '7'
    VK_THREE        = '3'; VK_EIGHT = '8'
    VK_FOUR         = '4'; VK_NINE  = '9'

    # Punctuation
    VK_BACKTICK     = '`'; VK_OPEN_PARENTHESIS  = '('; VK_PIPE              = '|'
    VK_TILDE        = '~'; VK_CLOSE_PARENTHESIS = ')'; VK_BACK_SLASH        = '\\'
    VK_EXCL_MARK    = '!'; VK_MINUS             = '-'; VK_BACK_COLON        = ':'
    VK_AT           = '@'; VK_UNDERSCORE        = '_'; VK_BACK_SEMI_COLON   = ';'
    VK_SHARP        = '#'; VK_PLUS              = '+'; VK_QUOTE             = '\''
    VK_DOLLAR       = '$'; VK_EQUALS            = '='; VK_DOUBLE_QUOTE      = '"'
    VK_PERCENTAGE   = '%'; VK_LEFT_BRACKET      = '['; VK_LOWER             = '<'
    VK_CIRCUMFLEX   = '^'; VK_RIGHT_BRACKET     = ']'; VK_GREATER           = '>'
    VK_AMPERSAND    = '&'; VK_LEFT_BRACE        = '{'; VK_COMMA             = ','
    VK_ASTERISK     = '*'; VK_RIGHT_BRACE       = '}'; VK_PERIOD            = '.'
    VK_SLASH        = '/'; VK_QUESTION_MARK     = '?'

    # fmt: on

    @staticmethod
    def kbhit(wait: bool = False, timeout: float = 0) -> bool:
        """Return whether there was an ENTER keyboard press."""
        while dr_list := select.select([sys.stdin], [], [], timeout):
            if not wait:
                return dr_list[0] != []
            elif dr_list[0]:
                return True

    @classmethod
    def getch(cls, n: int = 1) -> List["Keyboard"]:
        """Read and return (a) character(s) from a keyboard press.
        Method will return after ENTER is pressed.
        """
        return list(map(cls.of_value, sys.stdin.read(n)))

    @classmethod
    def digits(cls) -> List['Keyboard']:
        """Return all Keyboard digits."""
        return list(map(cls.of_value, filter(lambda v: str(v).isdigit(), cls.values())))

    @classmethod
    def letters(cls) -> List['Keyboard']:
        """Return all Keyboard letters."""
        return list(map(cls.of_value, filter(lambda v: str(v).isalpha(), cls.values())))

    @classmethod
    def wait_keystroke(cls, blocking: bool = True, ignore_error_keys: bool = True) -> Optional["Keyboard"]:
        """Wait until a keypress is detected."""
        try:
            keystroke = getkey.getkey(blocking)
            if keystroke:
                return cls.of_value(keystroke)
        except (KeyboardInterrupt, AssertionError) as err:
            if not ignore_error_keys:
                raise KeyboardInputError(f"Invalid keystroke => {str(err)}") from err
        except termios.error as err:
            raise NotATerminalError("keyboard:: Requires a terminal (TTY)") from err
        finally:
            sys.stdin.flush()

        return None

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return str(self)

    def isdigit(self) -> bool:
        """TODO"""
        return str(self.value).isdigit()

    def isalpha(self) -> bool:
        """TODO"""
        return str(self.value).isalpha()

    def isalnum(self) -> bool:
        """TODO"""
        return str(self.value).isalnum()

    def ispunct(self) -> bool:
        """TODO"""
        return all(ch in string.punctuation for ch in str(self.value))

    @property
    def val(self) -> str:
        return str(self.value)


if __name__ == '__main__':
    print(Keyboard.getch(3))
